import { defineStore } from 'pinia'
import { v4 } from 'uuid'
import {
  getProjectReferentials,
  getReferential,
  patchReferential, patchScope,
  postReferential,
} from '@/services/referentialsService'
import { getPaginatedResult } from './helpers'
import { notifyError } from '@/services/alertService'
import langIso from '@/assets/lang.json'
import type { PaginationObject } from '#/crud'
import type {
  Node,
  ReferentialState,
  ProjectReferential,
  ReferentialFilters,
  Translation,
  Hierarchy, Referential, LanguageDefinition,
} from '#/referential'


const initRef = (): ReferentialState => ({
  langIso,
  codeToLang: Object.fromEntries(langIso.map((lang: LanguageDefinition) => [lang.code, lang])),
  referential: {
    name: '',
    description: '',
    definition: {
      levels: [],
      nodes: {},
      roots: [],
    },
    languages: {
      available: [],
      default: null,
      translations: {},
    },
    scope: SCOPE_PROJECT,
  },
  projectReferentials: [],
  projectReferentialsCount: 0,
  pageCount: 1,
})

export const SCOPE_PROJECT = 4
export const SCOPE_PUBLIC = 2

export const useReferentialStore = defineStore('Referential', {
  state: () => initRef(),
  getters: {
    availableLanguages: (state): LanguageDefinition[] => state.referential.languages.available.map(code => state.codeToLang[code]),
    availableLanguagesStr: (state): string[] => state.referential.languages.available,
    defaultLanguage: (state): string | null => state.referential.languages.default,
    levelCount: (state): number => state.referential.definition.levels.length,
    isShared: (state): boolean => state.referential.scope < 4,
    levelFromId: (state) => (levelId: string): Hierarchy | undefined => state.referential.definition.levels.find(l => l.id === levelId) ,
    levels: (state): Hierarchy[] => state.referential.definition.levels,
    nodes: (state): Record<string, Node> => state.referential.definition.nodes,
    rootLevelName: (state): string => {
      const levels = state.referential.definition.levels
      return levels.length ? levels[0].name : ''
    },
    rootNodes: (state): Node[] => state.referential.definition.roots.map(id => state.referential.definition.nodes[id]),
    getChildren: (state) => (nodeId: string): string[] => {
      const node = state.referential.definition.nodes[nodeId]
      return node.children
          ? node.children
          : []
    },
    getTranslation: (state) => (id: string): Translation => state.referential.languages.translations[id],
    isTranslated: (state): boolean => Boolean(state.referential.languages.available.length)
  },
  actions: {
    levelAdd(name: string) {
      this.referential.definition.levels.push({name, id: v4()})
    },
    levelDelete(levelId: string) {
      const levelIndex = this.referential.definition.levels.findIndex(level => level.id === levelId)
      if (levelIndex === 0) {
        const newRoots: string[] = []
        for (const rootId of this.referential.definition.roots) {
          const {children} = this.referential.definition.nodes[rootId]
          if (children?.length) {
            children.forEach(childId => {
              const child = this.referential.definition.nodes[childId]
              if (child.level === 1) {
                newRoots.push(childId)
                delete child.ancestors
              } else {
                this.nodeDelete(childId, false)
              }
            })
          }
          delete this.referential.definition.nodes[rootId]
        }
        for (const node of Object.values(this.nodes)) {
          node.level --
        }
        this.referential.definition.roots = newRoots
        this.referential.definition.levels = this.levels.slice(1)
      }
      else if (levelIndex > 0) {
        this.referential.definition.levels.splice(levelIndex, 1)
        const cleanNode = (nodeId: string) => {
          const node = this.referential.definition.nodes[nodeId]
          if (node.level === levelIndex) { // give its children to its ancestors before removing node
            (node.ancestors as string[]).forEach(ancestorId => {
              const ancestor = this.referential.definition.nodes[ancestorId] as Required<Node>
              const childrenIds = node.children || []
              ancestor.children.splice(ancestor.children.indexOf(node.id), 1, ...childrenIds)

              childrenIds.forEach(childId => {
                const child = this.referential.definition.nodes[childId] as Required<Node>
                child.ancestors.splice(child.ancestors?.indexOf(nodeId), 1 , ancestorId)
              })
            })
            delete this.referential.definition.nodes[nodeId]
          }
          else if (node.level < levelIndex && node.children?.length) {
            for (const childId of [...node.children]) {
              cleanNode(childId)
            }
          }
        }
        this.referential.definition.roots.forEach(rootId => {
          const root = this.referential.definition.nodes[rootId]
          if (root.children) {
            root.children.forEach(childId => {
              cleanNode(childId)
            })
          }
        })
        for (const node of Object.values(this.referential.definition.nodes)) {
          if (node.level > levelIndex) {
            node.level--
          }
        }
      }
    },
    levelEdit(levelId: string, patch: Partial<Hierarchy>) {
      const levelIndex = this.referential.definition.levels.findIndex(l => l.id === levelId)
      const updatedLevel = this.referential.definition.levels[levelIndex]
      this.referential.definition.levels[levelIndex] = {...updatedLevel, ...patch}
    },
    nodeAdd(name: string, level = 0, ancestorId?: string) {
      const newNode = this.nodeCreate(name, level)
      if (ancestorId) {
        newNode.ancestors = [ancestorId]
        this.referential.definition.nodes[ancestorId].children = [
          ...this.referential.definition.nodes[ancestorId].children ?? [],
          newNode.id
        ]
      }
      else {
        this.referential.definition.roots.push(newNode.id)
      }
    },
    nodeCreate(name: string, level: number): Node {
      const nodeId = v4()
      const newNode = {
        id: nodeId,
        name,
        level,
      };
      this.referential.definition.nodes[nodeId] = newNode
      return newNode
    },
    nodeDelete(nodeId: string, ancestorReview = true) {
      const { children, ancestors, level } = this.referential.definition.nodes[nodeId]
      delete this.referential.definition.nodes[nodeId]
      if (level === 0) {
        const roots = this.referential.definition.roots
        roots.splice(roots.indexOf(nodeId), 1)
      } else if (ancestorReview && ancestors?.length) {
        ancestors.forEach(id => {
          const ancestor = this.nodes[id] as Required<Node>
          ancestor.children.splice(ancestor.children.indexOf(nodeId), 1)
        })
      }
      if (children?.length) {
        children.forEach(childId => {
          this.nodeDelete(childId, false)
        })
      }
    },
    nodeEdit(nodeId: string, patch: Partial<Node>) {
      const updatedBranch = this.referential.definition.nodes[nodeId]
      this.referential.definition.nodes[nodeId] = { ...updatedBranch, ...patch }
    },

    async referentialCreate(projectId: string) {
      this.referential = await postReferential(
          {
            projectId,
            ...this.referential,
          }
      )
    },
    async referentialLoad(referentialId: string) {
      this.referential = await getReferential(referentialId)
    },
    referentialUpdate() {
      const {id, name, description, definition, project , languages} = this.referential as Referential
      const projectId = project?.id
      return patchReferential({
        id, name, description, definition, projectId, languages
      })
    },
    async referentialShare(scopeData: {scope: 1 | 2 | 4}) {
      this.referential.scope = await patchScope((this.referential as Referential).id, scopeData)
    },
    setDefaultLanguage(newDefault: string | null) {
      this.referential.languages.default = newDefault
    },
    async loadProjectReferentials(pagination: PaginationObject, filter: ReferentialFilters, projectId: string): Promise<number> {
      return getProjectReferentials(pagination, filter, projectId)
       .then((paginatedReferentials) => {
         const { pageContent: referentials, pageCount, total } = getPaginatedResult<ProjectReferential>(paginatedReferentials, pagination.page, pagination.rowsPerPage)
         this.projectReferentials = referentials
         this.pageCount = pageCount
         this.projectReferentialsCount = total
         return total
       })
       .catch((error) => {
         notifyError(this.i18n.t('notify.error.set.all'), error)
        return 0
       })
    },
    updateTranslations(id: string, newTranslation: Translation) {
      this.referential.languages.translations[id] = newTranslation
    }
  }
})
